@@ -335,12 +335,9 @@ struct ui_form_opt {
 };
 
 #ifdef HAVE_ENGINE
- /* Ick. But there is no way to pass this sanely through OpenSSL */
-static struct openconnect_info *ui_vpninfo;
-
 static int ui_open(UI *ui)
 {
-	struct openconnect_info *vpninfo = ui_vpninfo; /* Ick */
+	struct openconnect_info *vpninfo = UI_get0_user_data(ui);
 	struct ui_data *ui_data;
 
 	if (!vpninfo || !vpninfo->process_auth_form)
@@ -354,6 +351,7 @@ static int ui_open(UI *ui)
 	ui_data->last_opt = &ui_data->form.opts;
 	ui_data->vpninfo = vpninfo;
 	ui_data->form.auth_id = (char *)"openssl_ui";
+
 	UI_add_user_data(ui, ui_data);
 
 	return 1;
@@ -433,42 +431,10 @@ static int ui_close(UI *ui)
 	return 1;
 }
 
-static UI_METHOD *create_openssl_ui(struct openconnect_info *vpninfo)
+static UI_METHOD *create_openssl_ui(void)
 {
 	UI_METHOD *ui_method = UI_create_method((char *)"AnyConnect VPN UI");
 
-	/* There is a race condition here because of the use of the
-	   static ui_vpninfo pointer. This sucks, but it's OpenSSL's
-	   fault and in practice it's *never* going to hurt us.
-
-	   This UI is only used for loading certificates from a TPM; for
-	   PKCS#12 and PEM files we hook the passphrase request differently.
-	   The ui_vpninfo variable is set here, and is used from ui_open()
-	   when the TPM ENGINE decides it needs to ask the user for a PIN.
-
-	   The race condition exists because theoretically, there
-	   could be more than one thread using libopenconnect and
-	   trying to authenticate to a VPN server, within the *same*
-	   process. And if *both* are using certificates from the TPM,
-	   and *both* manage to be within that short window of time
-	   between setting ui_vpninfo and invoking ui_open() to fetch
-	   the PIN, then one connection's ->process_auth_form() could
-	   get a PIN request for the *other* connection.
-
-	   However, the only thing that ever does run libopenconnect more
-	   than once from the same process is KDE's NetworkManager support,
-	   and NetworkManager doesn't *support* having more than one VPN
-	   connected anyway, so first that would have to be fixed and then
-	   you'd have to connect to two VPNs simultaneously by clicking
-	   'connect' on both at *exactly* the same time and then getting
-	   *really* unlucky.
-
-	   Oh, and the KDE support won't be using OpenSSL anyway because of
-	   licensing conflicts... so although this sucks, I'm not going to
-	   lose sleep over it.
-	*/
-	ui_vpninfo = vpninfo;
-
 	/* Set up a UI method of our own for password/passphrase requests */
 	UI_method_set_opener(ui_method, ui_open);
 	UI_method_set_writer(ui_method, ui_write);
@@ -498,12 +464,12 @@ static int pem_pw_cb(char *buf, int len, int w, void *v)
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("PEM password too long (%d >= %d)\n"),
 			     plen, len);
-		free(pass);
+		free_pass(&pass);
 		return -1;
 	}
 
 	memcpy(buf, pass, plen+1);
-	free(pass);
+	free_pass(&pass);
 	return plen;
 }
 
@@ -566,7 +532,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 			if (pass)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to decrypt PKCS#12 certificate file\n"));
-			free(pass);
+			free_pass(&pass);
 			if (request_passphrase(vpninfo, "openconnect_pkcs12", &pass,
 					       _("Enter PKCS#12 pass phrase:")) < 0) {
 				PKCS12_free(p12);
@@ -581,10 +547,10 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Parse PKCS#12 failed (see above errors)\n"));
 		PKCS12_free(p12);
-		free(pass);
+		free_pass(&pass);
 		return -EINVAL;
 	}
-	free(pass);
+	free_pass(&pass);
 	if (cert) {
 		char buf[200];
 		vpninfo->cert_x509 = cert;
@@ -615,7 +581,8 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 }
 
 #ifdef HAVE_ENGINE
-static int load_tpm_certificate(struct openconnect_info *vpninfo)
+static int load_tpm_certificate(struct openconnect_info *vpninfo,
+				const char *engine)
 {
 	ENGINE *e;
 	EVP_PKEY *key;
@@ -624,7 +591,11 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 
 	ENGINE_load_builtin_engines();
 
-	e = ENGINE_by_id("tpm");
+	e = ENGINE_by_id(engine);
+	if (!e && !strcmp(engine, "tpm2")) {
+		ERR_clear_error();
+		e = ENGINE_by_id("tpm2tss");
+	}
 	if (!e) {
 		vpn_progress(vpninfo, PRG_ERR, _("Can't load TPM engine.\n"));
 		openconnect_report_ssl_errors(vpninfo);
@@ -645,13 +616,13 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 				     _("Failed to set TPM SRK password\n"));
 			openconnect_report_ssl_errors(vpninfo);
 		}
-		vpninfo->cert_password = NULL;
-		free(vpninfo->cert_password);
-	} else {
-		/* Provide our own UI method to handle the PIN callback. */
-		meth = create_openssl_ui(vpninfo);
+		free_pass(&vpninfo->cert_password);
 	}
-	key = ENGINE_load_private_key(e, vpninfo->sslkey, meth, NULL);
+
+	/* Provide our own UI method to handle the PIN callback. */
+	meth = create_openssl_ui();
+
+	key = ENGINE_load_private_key(e, vpninfo->sslkey, meth, vpninfo);
 	if (meth)
 		UI_destroy_method(meth);
 	if (!key) {
@@ -673,7 +644,8 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 	return ret;
 }
 #else
-static int load_tpm_certificate(struct openconnect_info *vpninfo)
+static int load_tpm_certificate(struct openconnect_info *vpninfo,
+				const char *engine)
 {
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("This version of OpenConnect was built without TPM support\n"));
@@ -946,7 +918,11 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	while (fgets(buf, 255, f)) {
 		if (!strcmp(buf, "-----BEGIN TSS KEY BLOB-----\n")) {
 			fclose(f);
-			return load_tpm_certificate(vpninfo);
+			return load_tpm_certificate(vpninfo, "tpm");
+		} else if (!strcmp(buf, "-----BEGIN TSS2 KEY BLOB-----\n") ||
+			   !strcmp(buf, "-----BEGIN TSS2 PRIVATE KEY-----\n")) {
+			fclose(f);
+			return load_tpm_certificate(vpninfo, "tpm2");
 		} else if (!strcmp(buf, "-----BEGIN RSA PRIVATE KEY-----\n") ||
 			   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n") ||
 			   !strcmp(buf, "-----BEGIN EC PRIVATE KEY-----\n") ||
@@ -1024,7 +1000,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					if (pass) {
 						vpn_progress(vpninfo, PRG_ERR,
 							     _("Failed to decrypt PKCS#8 certificate file\n"));
-						free(pass);
+						free_pass(&pass);
 						pass = NULL;
 					}
 
@@ -1037,13 +1013,13 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					openconnect_report_ssl_errors(vpninfo);
 				}
 
-				free(pass);
+				free_pass(&pass);
 				vpninfo->cert_password = NULL;
 
 				X509_SIG_free(p8);
 				return -EINVAL;
 			}
-			free(pass);
+			free_pass(&pass);
 			vpninfo->cert_password = NULL;
 
 			key = EVP_PKCS82PKEY(p8inf);
@@ -1099,8 +1075,6 @@ int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
 
 static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 {
-	unsigned char sha256_hash[SHA256_SIZE];
-	unsigned char sha1_hash[SHA1_SIZE];
 	EVP_PKEY *pkey;
 	BIO *bp = BIO_new(BIO_s_mem());
 	BUF_MEM *keyinfo;
@@ -1119,14 +1093,11 @@ static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 
 	BIO_get_mem_ptr(bp, &keyinfo);
 
-	openconnect_sha256(sha256_hash, keyinfo->data, keyinfo->length);
-	openconnect_sha1(sha1_hash, keyinfo->data, keyinfo->length);
+	openconnect_sha256(vpninfo->peer_cert_sha256_raw, keyinfo->data, keyinfo->length);
+	openconnect_sha1(vpninfo->peer_cert_sha1_raw, keyinfo->data, keyinfo->length);
 
 	BIO_free(bp);
 
-	vpninfo->peer_cert_sha1 = openconnect_bin2hex("sha1:", sha1_hash, sizeof(sha1_hash));
-	vpninfo->peer_cert_sha256 = openconnect_bin2hex("sha256:", sha256_hash, sizeof(sha256_hash));
-
 	return 0;
 }
 
@@ -1658,10 +1629,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		X509_free(vpninfo->peer_cert);
 		vpninfo->peer_cert = NULL;
 	}
-	free (vpninfo->peer_cert_sha1);
-	vpninfo->peer_cert_sha1 = NULL;
-	free (vpninfo->peer_cert_sha256);
-	vpninfo->peer_cert_sha256 = NULL;
+	free(vpninfo->peer_cert_hash);
+	vpninfo->peer_cert_hash = NULL;
 	vpninfo->cstp_cipher = NULL;
 
 	ssl_sock = connect_https_socket(vpninfo);
